;; @(#) test-defsubst.el  -- Test the effect of defsubst
;;
;;  File id
;;
;;      $Docid: Jun 07 1997 15:38 Jari Aalto $
;;      $Contactid: <jari.aalto@poboxes.com> $
;;      $Id: test-defsubst.el,v 1.2 1997/06/07 12:38:23 jaalto Exp $
;;
;;  The interactive spec and defsubst suicide
;;
;;        [Moral: do no defsubst interactive functions]
;;        [Test suite by Steven L Baur <steve@calag.com> (XEmacs maintainer)]
;;
;;      Suppose you have some very small user callable function (a front
;;      end) and you wondered if you could just squeeze a little
;;      bit performance, if you changed the function spec from
;;
;;          defun  --> defsubst
;;
;;      Which would mean that, when function is compiled, it will be
;;      open code to the point where it was called. Let's start with
;;      this pseudo explanation. Here are the "lib" functions.
;;      The terms IACT here means that function has some interactive
;;      spec and (i-p) says that is uses (if (interactive-p) ...)
;;
;;          DEFUN lib-front-end
;;             IACT
;;             (i-p) (message "font-end")
;;             CALL main
;;             ..no more code..
;;
;;          DEFUN lib-main
;;              IACT
;;              (i-p) (message "main")
;;              ...100.. lines of code
;;
;;      Now, if have another function that uses front-end to access main
;;      (because front end is easy to call), he does this
;;
;;          DEFUN caller
;;             IACT
;;             (i-p) (message "caller)
;;             CALL lib-front-end
;;
;;      That creates function call chain
;;
;;          caller -> lib-front-end -> lib-main
;;
;;      But suppose you wanted to avoid the expense of function call of
;;      `lib-front-end', because it is so small function, and convert
;;      it into `defsubst'. After that the byte compiled call chain
;;      would look like this. There is only one function call left.
;;
;;          caller + (open-coded)lib-font-end --> lib-main.
;;
;;      Oh no, you shouldn't do that, because `lib-font-end' has IACT
;;      specs. If you do make it `defsubst', see what happen next. If
;;      you load the .el file, everything works the same, whether the
;;      `lib-font-end' was declared defsubst or not, the call
;;
;;          M-x caller prints messages only from `caller'
;;
;;      But if you byte compile file and load .elc with the `lib-front-end'
;;      being defsubst, then the call
;;
;;          M-x caller print messages from `caller' and `lib-font-end'
;;
;;      Surprise. The .el file and the .elc files behave totally differently.
;;      Here is what happened after byte compilation
;;
;;          DEFUN caller
;;            IACT
;;            (i-p)  (message-caller)
;;            [lib-font-end (i-p) (message "font-end")  ]
;;            [CALL main                                ]
;;
;;      I don't think you meant to inline `lib-font-end' (i-p) behaviour
;;      into `caller' function? This is the warning; in a nutshell.
;;
;;  Function names in this file
;;
;;      The names may look cryptic, but it is easier to follow if
;;      you can see from the name what the function was alike.
;;
;;          ip          ;defsubst without IACT and with (Interactive-p)
;;          IPip        ;defsubst with IACT and with (interactive-p)
;;          defun-nil   ;defun without IACT spec
;;          defun-P     ;defun with IACT spec
;;
;;      For example the following mean that there is function defun-P which
;;      has inlined defsubst IPip.
;;
;;          defun-P-IPip
;;
;;  Output of this module without compilation
;;
;;      Start emacs with emacs -q -l test.el.
;;      Go to (19.30+) *Messages* buffer or XEmacs 19.14 " *Messages"
;;      and see what the file printed. Notice the last calls, where
;;      the inlining is not happened yet. That when you call M-x.
;;
;;	    ip		    is _not_ interactive
;;	    defun-nil-IPip  is _not_ interactive
;;	    defun-P-ip	    is interactive
;;	    defun-P-IPip    is interactive
;;
;;          Calling interactively:
;;	    defun-P-ip:	    nil
;;	    defun-P-IPip:   nil
;;
;;  Output of this module _with_ compilation
;;
;;      Byte compile this file and start emacs with emacs -q -l test.elc.
;;      The message buffer will display followin. Now you see how
;;      the defsubst cause original defun's to behave differently.
;;
;;	    ip		    is _not_ interactive
;;	    defun-nil-IPip  is _not_ interactive
;;	    defun-P-ip	    is interactive
;;	    defun-P-IPip    is interactive
;;
;;          Calling interactively:
;;          Gotcha ip
;;	    defun-P-ip:	    "Gotcha ip"
;;          Gotcha IPip
;;	    defun-P-IPip:   "Gotcha IPip"

;;; ...................................... inlined function definition ...
;;;

(eval-when-compile (require 'cl))

(defsubst IPip (&optional arg)          ; I-P (i-p)
  (interactive "P")
  (when (interactive-p)
    (message "Gotcha IPip") "Gotcha IPip"))

(defsubst ip (&optional arg)            ; (i-p)
  (when (interactive-p)
      (message "Gotcha ip") "Gotcha ip"))

;;; ............................................................ defun ...

(defun defun-nil-IPip (&optional arg)
  (IPip))                               ;defsubst I-P (i-p)

(defun defun-P-ip (&optional arg)
  (interactive "P")
  (ip))                                 ;defsubst  (i-p)


(defun defun-P-IPip (&optional arg)
  (interactive "P")
  (IPip))                               ;defsubst  I-P (i-p)

;;; ............................................................. test ...



(when (commandp 'IPip)                  ;defsubst I-P (i-p)
  (message "IPip is interactive"))

(when (not (commandp 'ip))              ;defsubst (i-p)
  (message "ip is _not_ interactive"))

(when (not (commandp 'defun-nil-IPip))
  (message "defun-nil-IPip is _not_ interactive"))

(when (commandp 'defun-P-ip)            ;defun I-P + defsubst (i-p)
  (message "defun-P-ip is interactive"))

(when (commandp 'defun-P-IPip)          ;defun I-P + defsubst (i-p)
  (message "defun-P-IPip is interactive"))


(message "\nCalling interactively:")

(message "defun-P-ip: %s"   ;; --> "defun-P-ip: nil"
         (prin1-to-string (call-interactively 'defun-P-ip)))

(message "defun-P-IPip: %s"
         (prin1-to-string (call-interactively 'defun-P-IPip)))

;;; test-defsubst.el ends here
