;; @(#) tinychist.el -- Automatic command-history save/restore utility

;; @(#) $Id: tinychist.el,v 1.15 1997/05/31 11:50:32 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;; $KnownCompatibility: 19.28,29,30 XE 19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs


;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Apr 16 1996
;;
;; To get information on this program use ident(1) or do M-x tich-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinychist|Jari Aalto|jari.aalto@poboxes.com|
;; Automatic command-history save/restore utility between emacs sessions|
;; 31-May-1997|$Revision: 1.15 $|~/misc/tinychist.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;;  Put this file on your Emacs-Lisp load path, add following into your
;;  ~/.emacs startup file.
;;
;;	(require 'tinychist)
;;
;;  This package also installs function tich-save into kill-emacs-hook,
;;  or if no such hook exist, you will be warned that other means must
;;  be used.
;;
;;  Now, whenever you start emacs, name it with this option.
;;
;;	-name <TITLE>		, eg. -name Mail
;;
;;  and the history file associated with "Mail emacs" is loaded
;;
;;  If you have any questions, use this function
;;
;;      M-x tich-submit-bug-report


;;}}}
;;{{{ Docs

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tich-" in front of
;;   every function & variable. It stands for '(ti)ny (c)ommand (h)istory
;;   handler
;; - variable names contain letters 'tich-:'

;;  Preface, apr 1996
;;
;;	In newsgroup post to g.e.help there was discussion
;;	about saving and resatoring emacs command history between session.
;;
;;	Fred G. Athearn <fga@maple.sover.net> , gnu.emacs.help
;;	Send a help message to a person asking for it describing how to
;;	print out command-history and saving it into a file with C-x C-w.
;;
;;	This little package tries to automate everythig, so that when you
;;	load it, it will automatically load command history for the right
;;	emacs and when yu exit emacs, the command history is saved to disk.
;;
;;  Overview of features
;;
;;	o   Saving and restoring emacs command history
;;	o   Suports simultaneous emacs sessions, different history for
;;	    each one. Eg. "-name Mail" is associated with "mail" emacs history,
;;	    "-name News" is associated with "news" history. This trick works
;;	    on non-windowed tty too, since the switch is evaluated and cached
;;	    internally in those cases.
;;
;;  Default save file from -name parameter
;;
;;	The default save name of command history file is extracted from
;;	the frame parameter. It is quite customary that people have
;;	several emacs open in their X display, each one dedicated to
;;	specific task.
;;
;;	The key here is, that you should make a habbit of naming your
;;	emacs by task when you start it:
;;
;;	    % emacs -name Mail &	# My mail emacs
;;	    % emacs -name C++1 &	# My C++ project 1
;;	    % emacs -name News	# for news reading
;;
;;	This effectively sets the frame's name to "-name" parameter's value.
;;
;;	But old emacs  is a  little picky about the order of command line
;;	options, please loak at the info pages in which order you must specify
;;	additional arguments. (info pages, Node:Initial Options)
;;
;;	For non-windowed environment, this trick doesn't quite work out of
;;	the box, because emacs doesn't accept the name option at all.
;;	Let's try to start fresh emacs to an xterm, not to separate frame
;;	and see what happens. Order of the options is important here.
;;
;;	    % emacs -nw -q -name Mail  &
;;
;;	What happens, is that you get two new buffers: "-name" and "Mail", and
;;	this is not what we intended. If we ask the frame name in this
;;	emacs, it says "terminal" or something similar.
;;
;;	What we do instead, is ,that we install our own command line handler
;;	in non-windowed emacs and then we're able to intercept the "-name"
;;	option and it's parameter. When the emacs is killed, we then again
;;	look at the cached "-name" option to derive the save file postfix.
;;
;;	If you're interested in adding your own command line option,
;;	see function add-command-line-arg in tinylibm.el
;;
;;  How it works
;;
;;	Your emacs must support kill-emacs-hook, so that the command history
;;	can be saved automatically when you exit emacs. If you have older
;;	than 19.xx, or other emacs that doesn't support the variable,
;;	there is alternative example which replaces you emacs exit.
;;
;;	See at the end of file for examples.
;;
;;	When emacs loads, the tich-load is executed (see installation) and
;;	correct command-history file is loaded.
;;
;;  File saving/loading
;;
;;	If you want to save/load the history session manually, you can call
;;	function
;;
;;	    M-x tich-command-history-save-load
;;
;;  About using the frame name
;;
;;	I know it's a bit risky to rely on the frame name's first word, that
;;	it designates the purpose of your emacs. After all you _can_ change
;;	the frame name to whatever you want in side emacs.
;;
;;	But in general, the name labels your emacs and gives visible
;;	clue in X where the particular emacs is used to. (in non-X you can't
;;	see the name parameter at all, but I don't think that's a problem.)
;;
;;  See also
;;
;;	See file chistory.el in your emacs distribution, how to configure
;;	some parameters. Eg:
;;
;;	    M-x command-history-mode
;;	    (setq list-command-history-max 50)

;;}}}
;;{{{ History
;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May	31	1997	[jari]		19.28	v1.14		NotReleased
;; - Library function changes. Checkdoc 1.29 clear. Defcustom support.
;;
;; Apr	17	1997	[jari]		19.28	v1.12-1.13	NotReleased
;; - Cosmetic and internal changes.
;;
;; Jan	15	1997	[jari]		19.28	v1.9-1.11	NotReleased
;; - Rearranged loading. Now uses autoloads.
;; - Added lib m require due to macros.
;;
;; Sep	24	1996	[jari]		19.28	v1.8		NotReleased
;; - When saving the history file, it must now save entries that are not
;;   compatible for both emacs. Added re-incompatible to tich-save.
;;
;; Aug	22	1996	[jari]		19.28	v1.6-1.7	NotReleased
;; - Corrected all XE 19.14 ByteComp errors and made package XE compatible.
;;
;; May	16	1996	[jari]		19.28	v1.5		NotReleased
;; - Added extra ARG to tich-version function.
;;
;; Apr	16	1996	[jari]		19.28	v1.4		NotReleased
;; - Requires tinylibm.el 1.27
;; - Radical changes. Because the "-name" didn't work in non-windowed
;;   emacs, I had to code new command line handling.
;; - Also now history is loaded only once, successive calls to load-file
;;   tinychist.el does not reload history(wipeout effect).
;;
;; Apr	16	1996	[jari]		19.28	v1.2-1.3	NotReleased
;; - Some minor errors corrected. Tested in 19.30 too.
;;
;; Apr	16	1996	[jari]		19.28	v1.1		Created


;;}}}

;;; Code:

;;{{{ setup: require

(require 'chistory)
(require 'tinylibm)

(eval-when-compile (use-dynamic-compilation))

(defgroup-tiny TinyChist tich-: extensions
  "Automatic `command-history' save and restore utility.
  Overview of features

	o   Saving and restoring emacs command history
	o   Suports simultaneous emacs sessions, different history for
	    each one. Eg. '-name Mail' is associated with 'mail' emacs history,
	    '-name News' is associated with 'news' history. This trick works
	    on non-windowed tty too, since the switch is evaluated and cached
	    internally in those cases."
  )


;;}}}
;;{{{ setup: variables

;;; .......................................................... &v-bind ...

;;; ......................................................... &v-hooks ...

(defcustom tich-:load-hook nil
  "*Hook run when file is loaded."
  :type 'hook :group 'TinyChist
  )

;;; ....................................................... &v-private ...
;;; do not touch this

(defvar tich-:load-flag  nil
  "Non-nil means that flag is t when history has been loaded.")

;;; ........................................................ &v-public ...
;;; user configurable

;; I have a convention, where all my emacs startup files start with word
;; ~/.emacs. I have ~/.emacs.mail, ~/.emacs.kbd ~/.emacs.vc ~/.emacs.font
;; to name a few of my 30 startup files. Each one for specific task.
;;
(defcustom tich-:file-prefix  "~/.emacs.chist."
  "*File prefix for command history files."
  :type 'file :group 'TinyChist
  )

(defcustom tich-:non-x-function 'tich-non-x-name
  "*Function which return history file postfix in non-X emacs."
  :type 'function :group 'TinyChist
  )

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinychist.el"
 "tich"
 tich-:version-id
 "$Id: tinychist.el,v 1.15 1997/05/31 11:50:32 jaalto NotReleased $"
 '(tich-:version-id
   tich-:load-hook
   tich-:load-flag
   tich-:file-prefix
   tich-:non-x-function
   )))

;;}}}

;;; ########################################################### &Funcs ###

;;{{{ code: misc

;;; ----------------------------------------------------------------------
;;; - If you don't use -name in X, then the frame name looks like
;;;   EMACS-RUN-COMMAND@site.some.com
;;;
(defsubst tich-get-frame-word ()
  "Return first word of frame. "
  (let* ((name  (ti::xe-get-frame-parameter 'name))
         )
    (strmatget "^\\([^\t ]+\\)" 1 name)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tich-get-command-line-cache ()
  "Return option's -name ARG, if it was cached in non window env."
  (let* ((ptr  command-switch-alist)
	 arg

	 )
    ;; (("-name" . ignore) ("TITLE" . ignore) ..)
    ;;			   ^^^^^^^track down this
    (while ptr
      (if (not (string= "-name" (car (car ptr))))
	  (NEXTP ptr)
	(NEXTP ptr)			;get next assoc element.
	(setq arg (car-safe (car-safe ptr)))
	(setq ptr nil)			;stop there
	))
    arg
    ))

;;; ----------------------------------------------------------------------
;;; - This is just an sample, you propably want to modify this
;;;   to your env.
;;;
(defun tich-non-x-name ()
  "Return command history filename postfix.
Snoop what files are currently loaded into emacs. Eg. if RMAIL is found from
emacs, it is supposed that this emacs handles \"mail\"."
  (let* (elt
	 )

    ;; The order is important, because first one matched is used
    ;;
    (cond
     ((setq elt (tich-get-command-line-cache))
      ;; If this is non-window emacs, check if this optionis "cached" by
      ;; this package.
      elt
      )
     (;;  This is quite fast
      (ti::b-buffer-list
       '(string-match "VM$\\|RMAIL$\\|MH$" (buffer-name)) 'temp-buffers)
      "mail")

     ((ti::b-buffer-list
       '(string-match "gnus" (buffer-name)) 'temp-buffers)
      "news")

     (t
      "def"				;default
      ))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tich-get-file-name  ()
  "Return command history filename."
  (let* ((pfx    (expand-file-name tich-:file-prefix))
	 (frame  (tich-get-frame-word))
	 (func   tich-:non-x-function)
	 )
    (cond				;Is this non-X ?
     ((not (ti::xe-window-system))
      (setq frame (funcall func))))

    (concat pfx frame)
    ))

;;}}}
;;{{{ code: main

;;; ............................................................ &main ...

;;; ----------------------------------------------------------------------
;;; - I admit, this is a bit tricky thing to do, but it also demonstrates
;;;   how you'd add new cmd line options :-)
;;;
;;;      1. Install extra arg
;;;      2. read the arg content and and put it into list too
;;;
(defun tich-load  ()
  "Load appropriate `command-history' file by looking at cmd line option.
Optiona name seeked is `-name'. This function runs when package is loaded."
  (let* ((elt1			(member "-name" command-line-args))
	 (elt2			(car-safe (cdr elt1)))
	 (pfx			(expand-file-name tich-:file-prefix))
	 (debug-on-error	t)	;trigger errors
	 file
	 )
    (cond
     ((not elt1)			;no option given
      (setq file (tich-get-file-name)))	;guess which file to load.
     (t
      ;;  We know what user wants to load...
      ;;  Prevent using this arg as buffer name. The "-name"
      ;;  option is valid only in X
      ;;
      (cond
       ((not (ti::xe-window-system))
	(add-command-line-arg elt2)	  ;; this is put after
	(add-command-line-arg (car elt1)) ;; this is put before
	))

      (setq file (concat pfx elt2))
      ))

    (if (and file (file-exists-p file))
	(load file))

    ))

;;; ----------------------------------------------------------------------
;;;
(defun tich-save  ()
  "Load the saved `command-history' file."
  (let* ((file (tich-get-file-name))
	 )
    (and command-history		;something to save ?
	 (tich-command-history-save-load file))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tich-command-history-save-load  (file &optional load)
  "Save history to FILE or LOAD command history."
  (interactive "FFile: ")
  (let* ((verb           (interactive-p))
         (buffer         "*Command History*") ;see chistory.el
	 (debug-on-error t)		;trigger errors
	 (re-incompatible
	  "describe-key"		;XE and Emacs don't agree with this
	  )
         )

    (cond                               ;handle interactive
     (verb
      (cond
       ((y-or-n-p "Save history [or load] ?")
        (setq load nil))
       (t
        (setq load t)
        ))))

    (cond
     (load
      (load-file file))
     (t
      (list-command-history)
      (set-buffer buffer)
      (setq buffer-read-only nil)

      ;;  Format it
      (PMIN)
      (flush-lines re-incompatible)
      (PMIN)

      (insert
       (concat
	";;\n"
        ";; File: Emacs command history file\n"
        ";; Desc: Load this with M-x load-file to reset the history to "
        "these values\n"

	;;	This is for my tinystamp.el -- RCS compatible timestamp
	;;
        ";; $\Docid: $\n"
        "\n"
        ))

      (insert "(defconst command-history\n'(\n")

      (goto-char (point-max))
      (insert
       (concat
        "))\n\n"
        ";;; " (file-name-nondirectory file) " ends here"
        ))

      (lisp-mode)
      (indent-region (point-min) (point-max) nil)

      ;; Update the timestamp [if func exist] and save the configuration
      ;; tinystamp.el
      ;;
      (when (fboundp 'tis-save-buffer)
	(sfuncall 'tis-save-buffer))

      (write-region (point-min) (point-max) file)
      (kill-buffer buffer)
      ))
    ))

;;}}}
;;{{{ code: examples

;;; ......................................................... &example ...
;;; Code can be easily extracted with tinylib.el / ti::m-pkg-rip-magic
;;; This is Manual C-xC-x installation
;;;
;;; - we have to replace your kill emacs function with this, so that
;;;   the history is saved. Copy this into your emacs BEFORE the
;;;   (require 'tinychist)
;;; - Modify the function any way you like.

;;* _
;;* (global-set-key "\C-x\C-c" 'my-kill-emacs)
;;* _
;;* ;; This works for 18.57 and 19.xx, whereas the
;;* ;; kill-emacs-query-functions doesn't, and this suffices for personal
;;* ;; needs.
;;* _
;;* (defun my-kill-emacs (&optional arg)
;;*   "Prevent accidental emacs kill. Calls save-buffers-kill-emacs"
;;*   (interactive "P")
;;*   (let* ((chist (tich-get-file-name))
;;*          ch
;;*          )
;;*     (setq ch (read-char-safe-until "Exit emacs y/[n]"))
;;*     (cond
;;*      ((char= ch ?y)
;;*       (tich-command-history-save-load chist)
;;*       (save-buffers-kill-emacs arg)
;;*       ))))

;;}}}
;;{{{ code: install

;;; .................................................... &auto-install ...

(cond
 ((boundp 'kill-emacs-hook)
  (add-hook 'kill-emacs-hook 'tich-save))

 ;;  Default binding, this runs the kill emacs hook. We know that,
 ;;  but if user has relocated it then....
 ;;
 ((memq (lookup-key global-map "\C-x\C-c")
	'(save-buffers-kill-emacs))
  (read-from-minibuffer
   "Cannot auto-install. Use manual C-xC-c installation. See source code."
   )))


;; load only once, prevent command-history wipeout
;; if package is loaded multiple times
;;
(if (null tich-:load-flag)
    (tich-load)				;Autoload the command history
  (setq tich-:load-flag t))


(provide   'tinychist)
(run-hooks 'tich-:load-hook)

;;}}}


;;; tinychist.el ends here
