;; @(#) tinylpr.el -- Easy Emacs lpr command handling, popup, completions

;; @(#) $Id: tinylpr.el,v 1.10 1997/05/14 08:11:58 jaalto NotReleased $
;; @(#) $Keywords: tools, printer, lpr $
;; $KnownCompatibility: 19.28 $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Feb 16 1996
;;
;; To get information on this program use ident(1) or do M-x tilp-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinylpr|Jari Aalto|jari.aalto@poboxes.com|
;; Easy Emacs lpr command handling, popup, completions|
;; 14-May-1997|$Revision: 1.10 $|~/misc/tinylpr.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;;   Put this file on your Emacs-Lisp load path, add following into your
;;   ~/.emacs startup file
;;
;;	(require 'tinylpr)
;;
;;   Notice, that loading this file changes your lpr-command immediately
;;   to "sh". This is essential and if you want to use this package,
;;   leave it there or choose some compatible shell that accepts "-c"
;;   switch.
;;
;;
;;   CHANGE THE VARIABLES !
;;
;;   You must copy the user variables and put your own definitions
;;   there. The ones that ship with this module are only examples
;;
;;   If you have any questions, use this function
;;
;;      M-x tilp-submit-feedback

;;}}}
;;{{{ docs

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tilp-" in front of
;;   every function & variable. It stands for '(ti)ny (l)ine (pr)inter
;;   handler
;; - variable names contain letters 'tilp-:'

;;  Preface, 1996, overview of features
;;
;;	o   Managing printers or print styles easily
;;	o   Quesue information
;;	o   Has ready X-popup example to select print styles etc.
;;
;;  Story behind this package
;;
;;	Long time ago, I get frustrated that in out environment there existed
;;	dozen printers and also I had dozed ways to print the file. In
;;	HP-UX there exists many printing related commands:
;;
;;	    mmpage	,multi-page 1-8, like sunOS enscript
;;	    lp
;;	    lpr
;;	    a2ps
;;	    squeeze.awk	,my own empty line squeezer. Ever run CPP on C/C++ ?
;;	    groff	,I make some nroff files...
;;	    banner	,big letters
;;	    lpstat
;;	    col -bx	,remove ctrl codes from man pages
;;	    pps		,pretty printer for PostScript -- jau@tut.fi
;;	    pr		,format files
;;	    fold	,fold long lines for finite width output device
;;	    adjust	,for filling, centering, ..justifying
;;
;;	If I wanted to print the file in some other format, ie combining
;;	some of the commands I had to go and chnage the
;;
;;	    lpr-switches
;;
;;	every time. That was real tedious. And I stopped that after first
;;	experiments. Instead I made nice popup menu to emacs where I could
;;	see lpstatus, select print command, cancel print job etc...
;;
;;  Suggested keybindings
;;
;;* _
;;* (global-set-key "\C-z" nil)
;;* (global-set-key "\C-zpd"		'tilp-select-printer) ;; destin.
;;* (global-set-key "\C-zpq"		'tilp-queue)
;;* (global-set-key "\C-zps"		'tilp-print-style-select)
;;* (global-set-key "\C-zpr"		'print-region)
;;* (global-set-key "\C-zpb"		'print-buffer)
;;* _
;;* ;;  Se tinylib.el -- &examples section
;;* ;;
;;* (global-set-key "\C-zpn"		'my-print-with-line-numbers)
;;* _



;;}}}
;;{{{ History

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May	14	1997	[jari]		19.28	v1.9-1.10	NotReleased
;; - internal changes
;; - Changed all non standard boolean tests to use when and unless
;;
;; Jan	15	1997	[jari]		19.28	v1.7-1.8	NotReleased
;; - Levent N Atasoy <Levent.Atasoy@macsch.com> suggested that history
;;   list should be used in print styles and printers selected.
;; - Now also sets ps-lpr-switches if option
;;   tilp-:set-ps-lpr-switches is non-nil when new printer
;;   has been selected.
;;
;; Nov	1	1996	[jari]		19.28	v1.4-1.6	NotReleased
;; - Rearranged loading. Now uses autoloads.
;; - Changed all '(lambda forms to use (function (lambda
;; - Added lib m require due to macros.
;;
;; Apr	26	1996	[jari]		19.28	v1.3		NotReleased
;; - Textual changes
;;
;; Feb	16	1996	[jari]		19.28	v1.2		NotReleased
;; - Documentation fixes.
;;
;; Feb	16	1996	[jari]		19.28	v1.1		Created

;;}}}


;;; Code:

;;{{{ setup: require

(require 'tinylibm)

(eval-and-compile
  (defvar ps-lpr-switches nil)		;to shut up ByteCompiler
  )

(eval-when-compile (use-dynamic-compilation))

;;}}}
;;{{{ setup: variables

;;; .......................................................... &v-bind ...

;;; ......................................................... &v-hooks ...

(defvar tilp-:load-hook nil
  "Hook run when file is loaded.")

;;; .......................................................... &v-vars ...
;;; *** important ***
;;;
;;;     These are just examples. Copy the variables into your ~/.emacs
;;;     and make changes to reflect your system.
;;;
;;;

(defvar tilp-:set-ps-lpr-switches  t
  "If non-nil, set also ps-lpr-switches from ps-print.el when
changing printer.")


(defvar tilp-:queue-cmd "lpstat"
  "*Shell Command to return queue status")

(defvar tilp-:printer-list '("03c3" "03d3")
  "*List of available printers.")

(defvar tilp-:print-style-list
  ;;  - I have aliases the 'mmpage' , so use absolute command now
  ;;  - I need this let command due to local variables
  ;;
  ;;  - Select the first string so, that it's easy to complete.
  ;;    ie "p" TAB triggers the first choice
  ;;
  (let* ((mp   (or (getenv "MMPAGE")

;;; This is commented because this is example setting and loading
;;; the file would otherwise cause error, because you don't have
;;; this env variable.
;;;
;;;		   (error "define $MMPAGE env. var")
		   ))
	 styles
	 )
    (setq
     styles
     (list
      (list
       "p ,straight lp"		"lp -d#")
      (list
       "2 pages"		(concat  mp " -A -2 -P#"))
      (list
       "4 pages"		(concat  mp " -A -4 -P#"))
      (list
       "8 pages"		(concat  mp " -A -8 -P#"))
      (list
       "2l landscape"		(concat  mp " -A -l -2 -P#"))

      (list
       "q4 squeeze empty lines"
       (concat " txtlines.awks | " mp "-A -4 -P#"))

      (list
       "q8 squeeze empty lines"
       (concat " txtlines.awks | " mp "-A -8 -P#"))

      (list
       "lisp code printing"
       (concat "lispp.sh"))

      (list
       "C/C++ code printing"
       (concat "c-print.sh"))
      ))
    styles
    )
  "*Available print styles. The # char tells where to install printer in
command.

Format:  '((COMPLETION-STRING PRINTER-COMMAND-STRING) ..)

")


;;; ....................................................... &v-private ...

(defvar tilp-:current-printer (car  tilp-:printer-list)
  "Private. Current destination printer.")

(defvar tilp-:current-print-style (car (car  tilp-:print-style-list))
  "Private. Current print style.")

(defvar tilp-:printer-list-history nil
  "Private. History list for tilp-:printer-list")

(defvar tilp-:print-style-history nil
  "Private. History list for tilp-print-style-completions.")


;;; ....................................................... &v-version ...


(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinylpr.el"
 "tilp"
 tilp-:version-id
 "$Id: tinylpr.el,v 1.10 1997/05/14 08:11:58 jaalto NotReleased $"
 '(tilp-:version-id
   tilp-:load-hook
   tilp-:set-ps-lpr-switches
   tilp-:queue-cmd
   tilp-:printer-list
   tilp-:print-style-list
   tilp-:current-printer
   tilp-:current-print-style
   tilp-:printer-list-history
   tilp-:print-style-history
   )))


;;}}}

;;; ########################################################### &Funcs ###

;;{{{ code: funcs

;;; ----------------------------------------------------------------------
;;;
(defun tilp-set-command (template printer)
  "Substitutes possible # with PRINTER name in print command TEMPLATE."
  (if (string-match "\\(#\\)" template)
      (setq template (kill-match 1 printer template)))

  ;;  We know the lpr-command is "sh", so just put the "-c" as
  ;;  first option.
  ;;
  (setq lpr-switches (list "-c" template))
  )



;;; ----------------------------------------------------------------------
;;;
(defun tilp-print-style-completions ()
  ;;  Build up the completion array
  (let* ((list  tilp-:print-style-list)
	 (i 0)
	 completions
	 )
    (mapcar
     (function
      (lambda (x)
	(setq i (1+ i))
	(setq completions (cons  (cons (car x) i) completions))))
     list)
    completions
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tilp-setting-status ()
  "Returns current settings."
  (interactive)
  (let* ((stat	(nth 1 lpr-switches))
	 )
    (if (interactive-p)
	(message stat))
    stat
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tilp-queue ()
  "Returns queue status"
  (interactive)
  (let* ((cmd       tilp-:queue-cmd)
	 (buffer    (use-tmp-buffer "*tmp*" 'clear))
	 )
    (display-buffer buffer)
    (shell-execute cmd buffer)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tilp-select-printer (arg)
  "Select printer"
  (interactive
   (list
    (completing-read
     (concat "Printer [" tilp-:current-printer "]: ")
     (list-to-assoc-menu tilp-:printer-list)
     nil t
     nil
     'tilp-:printer-list-history
     )))

  (when (not (var-nil-p arg))
    (setq tilp-:current-printer arg)

    (if tilp-:set-ps-lpr-switches
	(setq ps-lpr-switches (list (concat "-P" arg))))

    (tilp-print-style-select tilp-:current-print-style)
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tilp-print-style-select (arg)
  "Seelct print output style."
  (interactive
   (list
    (completing-read
     (format "Print style [%s: %s ]: "
	     (or tilp-:current-print-style "<style unknown>")
	     (tilp-setting-status)
	     )
     (tilp-print-style-completions)
     nil t
     )))


  (let* ((printer (or tilp-:current-printer ""))
	 elt
	 args
	 )
    ;;  Try to find the style in assoc array
    ;;
    (if (not (and arg (setq elt (assoc arg tilp-:print-style-list))))
	(message "No such style")
      ;;  replace # with printer name
      ;;
      (setq tilp-:current-print-style arg)
      (setq args (nth 1 elt))
      (tilp-set-command args printer)
      (message "Print <%s> on %s" arg (tilp-setting-status))
      )))

;;}}}
;;{{{ examples

;;; ........................................................ &examples ...
;;; Here is how you could run commnds from emacs popup
;;; - Code can be easily extracted with tinylib.el / ti::m-pkg-rip-magic
;;;

;;* _
;;* (defun my-x-menu (event)
;;*   "Popups an X window of user defined commands. "
;;*   (interactive "e")
;;*   (let* ((pstat
;;* 	  (ti::s-substitute		;drep away directory name
;;* 	   (or (my-print-status) "")
;;* 	   ".*/"
;;* 	   ""
;;* 	   ))
;;* 	 item
;;* 	 )
;;*     (setq
;;*      item
;;*      (x-popup-menu
;;*       event
;;* _
;;*       (list
;;*        "Command Menu"
;;*        (list
;;* 	"Printer: "
;;* 	;;   This first one is header, not selection
;;* 	(cons (concat ":: " pstat) 'ignore)
;;* _
;;* 	;; these are selections
;;* 	'("* Print region"	. print-region)
;;* 	'("* Print buffer"	. print-buffer)
;;* 	'("Destination"		. tilp-select-printer)
;;* 	'("Print style"		. tilp-print-style-select)
;;* 	'("Queue status"	. tilp-queue)
;;* 	))
;;*       ))
;;* _
;;*     (cond
;;*      (item					;direct command
;;*       (call-interactively item)
;;*       ))
;;*     ))

;;}}}
;;{{{ code: auto install

;;; .................................................... &auto-install ...

;; Install package, reset lpr variables
;;
(cond
 ((not (string-match "sh" lpr-command))
  (setq lpr-command   "sh")
  (tilp-set-command (nth 1 (car  tilp-:print-style-list))
		    tilp-:current-printer)
  ))

;;}}}

(provide 'tinylpr)
(run-hooks 'tilp-:load-hook)


;;; tinylpr.el ends here
